#include <Python.h>


PyObject* method_EN_createproject(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_EN_deleteproject(PyObject* self, PyObject* args);
PyObject* method_EN_init(PyObject* self, PyObject* args);
PyObject* method_EN_open(PyObject* self, PyObject* args);
PyObject* method_EN_openX(PyObject* self, PyObject* args);
PyObject* method_EN_gettitle(PyObject* self, PyObject* args);
PyObject* method_EN_settitle(PyObject* self, PyObject* args);
PyObject* method_EN_getcomment(PyObject* self, PyObject* args);
PyObject* method_EN_setcomment(PyObject* self, PyObject* args);
PyObject* method_EN_getcount(PyObject* self, PyObject* args);
PyObject* method_EN_saveinpfile(PyObject* self, PyObject* args);
PyObject* method_EN_close(PyObject* self, PyObject* args);
PyObject* method_EN_solveH(PyObject* self, PyObject* args);
PyObject* method_EN_usehydfile(PyObject* self, PyObject* args);
PyObject* method_EN_openH(PyObject* self, PyObject* args);
PyObject* method_EN_initH(PyObject* self, PyObject* args);
PyObject* method_EN_runH(PyObject* self, PyObject* args);
PyObject* method_EN_nextH(PyObject* self, PyObject* args);
PyObject* method_EN_saveH(PyObject* self, PyObject* args);
PyObject* method_EN_savehydfile(PyObject* self, PyObject* args);
PyObject* method_EN_closeH(PyObject* self, PyObject* args);
PyObject* method_EN_solveQ(PyObject* self, PyObject* args);
PyObject* method_EN_openQ(PyObject* self, PyObject* args);
PyObject* method_EN_initQ(PyObject* self, PyObject* args);
PyObject* method_EN_runQ(PyObject* self, PyObject* args);
PyObject* method_EN_nextQ(PyObject* self, PyObject* args);
PyObject* method_EN_stepQ(PyObject* self, PyObject* args);
PyObject* method_EN_closeQ(PyObject* self, PyObject* args);
PyObject* method_EN_writeline(PyObject* self, PyObject* args);
PyObject* method_EN_report(PyObject* self, PyObject* args);
PyObject* method_EN_copyreport(PyObject* self, PyObject* args);
PyObject* method_EN_clearreport(PyObject* self, PyObject* args);
PyObject* method_EN_resetreport(PyObject* self, PyObject* args);
PyObject* method_EN_setreport(PyObject* self, PyObject* args);
PyObject* method_EN_setstatusreport(PyObject* self, PyObject* args);
PyObject* method_EN_getversion(PyObject* self, PyObject* args);
PyObject* method_EN_geterror(PyObject* self, PyObject* args);
PyObject* method_EN_getstatistic(PyObject* self, PyObject* args);
PyObject* method_EN_getresultindex(PyObject* self, PyObject* args);
PyObject* method_EN_getoption(PyObject* self, PyObject* args);
PyObject* method_EN_setoption(PyObject* self, PyObject* args);
PyObject* method_EN_getflowunits(PyObject* self, PyObject* args);
PyObject* method_EN_setflowunits(PyObject* self, PyObject* args);
PyObject* method_EN_gettimeparam(PyObject* self, PyObject* args);
PyObject* method_EN_settimeparam(PyObject* self, PyObject* args);
PyObject* method_EN_getqualinfo(PyObject* self, PyObject* args);
PyObject* method_EN_getqualtype(PyObject* self, PyObject* args);
PyObject* method_EN_setqualtype(PyObject* self, PyObject* args);
PyObject* method_EN_addnode(PyObject* self, PyObject* args);
PyObject* method_EN_deletenode(PyObject* self, PyObject* args);
PyObject* method_EN_getnodeindex(PyObject* self, PyObject* args);
PyObject* method_EN_getnodeid(PyObject* self, PyObject* args);
PyObject* method_EN_setnodeid(PyObject* self, PyObject* args);
PyObject* method_EN_getnodetype(PyObject* self, PyObject* args);
PyObject* method_EN_getnodevalue(PyObject* self, PyObject* args);
PyObject* method_EN_setnodevalue(PyObject* self, PyObject* args);
PyObject* method_EN_setjuncdata(PyObject* self, PyObject* args);
PyObject* method_EN_settankdata(PyObject* self, PyObject* args);
PyObject* method_EN_getcoord(PyObject* self, PyObject* args);
PyObject* method_EN_setcoord(PyObject* self, PyObject* args);
PyObject* method_EN_getdemandmodel(PyObject* self, PyObject* args);
PyObject* method_EN_setdemandmodel(PyObject* self, PyObject* args);
PyObject* method_EN_adddemand(PyObject* self, PyObject* args);
PyObject* method_EN_deletedemand(PyObject* self, PyObject* args);
PyObject* method_EN_getdemandindex(PyObject* self, PyObject* args);
PyObject* method_EN_getnumdemands(PyObject* self, PyObject* args);
PyObject* method_EN_getbasedemand(PyObject* self, PyObject* args);
PyObject* method_EN_setbasedemand(PyObject* self, PyObject* args);
PyObject* method_EN_getdemandpattern(PyObject* self, PyObject* args);
PyObject* method_EN_setdemandpattern(PyObject* self, PyObject* args);
PyObject* method_EN_getdemandname(PyObject* self, PyObject* args);
PyObject* method_EN_setdemandname(PyObject* self, PyObject* args);
PyObject* method_EN_addlink(PyObject* self, PyObject* args);
PyObject* method_EN_deletelink(PyObject* self, PyObject* args);
PyObject* method_EN_getlinkindex(PyObject* self, PyObject* args);
PyObject* method_EN_getlinkid(PyObject* self, PyObject* args);
PyObject* method_EN_setlinkid(PyObject* self, PyObject* args);
PyObject* method_EN_getlinktype(PyObject* self, PyObject* args);
PyObject* method_EN_setlinktype(PyObject* self, PyObject* args);
PyObject* method_EN_getlinknodes(PyObject* self, PyObject* args);
PyObject* method_EN_setlinknodes(PyObject* self, PyObject* args);
PyObject* method_EN_getlinkvalue(PyObject* self, PyObject* args);
PyObject* method_EN_setlinkvalue(PyObject* self, PyObject* args);
PyObject* method_EN_setpipedata(PyObject* self, PyObject* args);
PyObject* method_EN_getvertexcount(PyObject* self, PyObject* args);
PyObject* method_EN_getvertex(PyObject* self, PyObject* args);
PyObject* method_EN_setvertices(PyObject* self, PyObject* args);
PyObject* method_EN_getpumptype(PyObject* self, PyObject* args);
PyObject* method_EN_getheadcurveindex(PyObject* self, PyObject* args);
PyObject* method_EN_setheadcurveindex(PyObject* self, PyObject* args);
PyObject* method_EN_addpattern(PyObject* self, PyObject* args);
PyObject* method_EN_deletepattern(PyObject* self, PyObject* args);
PyObject* method_EN_getpatternindex(PyObject* self, PyObject* args);
PyObject* method_EN_getpatternid(PyObject* self, PyObject* args);
PyObject* method_EN_setpatternid(PyObject* self, PyObject* args);
PyObject* method_EN_getpatternlen(PyObject* self, PyObject* args);
PyObject* method_EN_getpatternvalue(PyObject* self, PyObject* args);
PyObject* method_EN_setpatternvalue(PyObject* self, PyObject* args);
PyObject* method_EN_getaveragepatternvalue(PyObject* self, PyObject* args);
PyObject* method_EN_setpattern(PyObject* self, PyObject* args);
PyObject* method_EN_addcurve(PyObject* self, PyObject* args);
PyObject* method_EN_deletecurve(PyObject* self, PyObject* args);
PyObject* method_EN_getcurveindex(PyObject* self, PyObject* args);
PyObject* method_EN_getcurveid(PyObject* self, PyObject* args);
PyObject* method_EN_setcurveid(PyObject* self, PyObject* args);
PyObject* method_EN_getcurvelen(PyObject* self, PyObject* args);
PyObject* method_EN_getcurvetype(PyObject* self, PyObject* args);
PyObject* method_EN_getcurvevalue(PyObject* self, PyObject* args);
PyObject* method_EN_setcurvevalue(PyObject* self, PyObject* args);
PyObject* method_EN_getcurve(PyObject* self, PyObject* args);
PyObject* method_EN_setcurve(PyObject* self, PyObject* args);
PyObject* method_EN_addcontrol(PyObject* self, PyObject* args);
PyObject* method_EN_deletecontrol(PyObject* self, PyObject* args);
PyObject* method_EN_getcontrol(PyObject* self, PyObject* args);
PyObject* method_EN_setcontrol(PyObject* self, PyObject* args);
PyObject* method_EN_addrule(PyObject* self, PyObject* args);
PyObject* method_EN_deleterule(PyObject* self, PyObject* args);
PyObject* method_EN_getrule(PyObject* self, PyObject* args);
PyObject* method_EN_getruleID(PyObject* self, PyObject* args);
PyObject* method_EN_getpremise(PyObject* self, PyObject* args);
PyObject* method_EN_setpremise(PyObject* self, PyObject* args);
PyObject* method_EN_setpremiseindex(PyObject* self, PyObject* args);
PyObject* method_EN_setpremisestatus(PyObject* self, PyObject* args);
PyObject* method_EN_setpremisevalue(PyObject* self, PyObject* args);
PyObject* method_EN_getthenaction(PyObject* self, PyObject* args);
PyObject* method_EN_setthenaction(PyObject* self, PyObject* args);
PyObject* method_EN_getelseaction(PyObject* self, PyObject* args);
PyObject* method_EN_setelseaction(PyObject* self, PyObject* args);
PyObject* method_EN_setrulepriority(PyObject* self, PyObject* args);
PyObject* method_EN_gettag(PyObject* self, PyObject* args);
PyObject* method_EN_settag(PyObject* self, PyObject* args);
PyObject* method_EN_timetonextevent(PyObject* self, PyObject* args);
PyObject* method_EN_getnodevalues(PyObject* self, PyObject* args);
PyObject* method_EN_getlinkvalues(PyObject* self, PyObject* args);
PyObject* method_EN_setvertex(PyObject* self, PyObject* args);
PyObject* method_EN_loadpatternfile(PyObject* self, PyObject* args);
PyObject* method_EN_setcurvetype(PyObject* self, PyObject* args);
PyObject* method_EN_getcontrolenabled(PyObject* self, PyObject* args);
PyObject* method_EN_setcontrolenabled(PyObject* self, PyObject* args);
PyObject* method_EN_getruleenabled(PyObject* self, PyObject* args);
PyObject* method_EN_setruleenabled(PyObject* self, PyObject* args);