LIBRARY     EPANET2.DLL

EXPORTS
    ENaddcontrol                  = _ENaddcontrol@24                    
    ENaddcurve                    = _ENaddcurve@4                     
    ENaddlink                     = _ENaddlink@20
    ENaddnode                     = _ENaddnode@12
    ENadddemand                   = _ENadddemand@16
    ENaddpattern                  = _ENaddpattern@4
    ENaddrule                     = _ENaddrule@4
    ENclearreport                 = _ENclearreport@0
    ENclose                       = _ENclose@0                          
    ENcloseH                      = _ENcloseH@0                         
    ENcloseQ                      = _ENcloseQ@0
    ENcopyreport                  = _ENcopyreport@4
    ENdeletecontrol               = _ENdeletecontrol@4
    ENdeletecurve                 = _ENdeletecurve@4
    ENdeletedemand                = _ENdeletedemand@8
    ENdeletelink                  = _ENdeletelink@8
    ENdeletenode                  = _ENdeletenode@8
    ENdeletepattern               = _ENdeletepattern@4
    ENdeleterule                  = _ENdeleterule@4
    ENepanet                      = _ENepanet@16
    ENgetaveragepatternvalue      = _ENgetaveragepatternvalue@8
    ENgetbasedemand               = _ENgetbasedemand@12
    ENgetcomment                  = _ENgetcomment@12
    ENgetcontrol                  = _ENgetcontrol@24                    
    ENgetcoord                    = _ENgetcoord@12
    ENgetcount                    = _ENgetcount@8                       
    ENgetcurve                    = _ENgetcurve@20
    ENgetcurveid                  = _ENgetcurveid@8                   
    ENgetcurveindex               = _ENgetcurveindex@8                
    ENgetcurvelen                 = _ENgetcurvelen@8                  
    ENgetcurvetype                = _ENgetcurvetype@8
    ENgetcurvevalue               = _ENgetcurvevalue@16 
    ENgetdemandindex              = _ENgetdemandindex@12
    ENgetdemandmodel              = _ENgetdemandmodel@16
    ENgetdemandname               = _ENgetdemandname@12
    ENgetdemandpattern            = _ENgetdemandpattern@12
    ENgetelseaction               = _ENgetelseaction@20
    ENgeterror                    = _ENgeterror@12                      
    ENgetflowunits                = _ENgetflowunits@4                   
    ENgetheadcurveindex           = _ENgetheadcurveindex@8
    ENgetlinkid                   = _ENgetlinkid@8                      
    ENgetlinkindex                = _ENgetlinkindex@8                   
    ENgetlinknodes                = _ENgetlinknodes@12                  
    ENsetlinknodes                = _ENsetlinknodes@12                  
    ENgetlinktype                 = _ENgetlinktype@8                    
    ENgetlinkvalue                = _ENgetlinkvalue@12
    ENgetnodeid                   = _ENgetnodeid@8                      
    ENgetnodeindex                = _ENgetnodeindex@8                   
    ENgetnodetype                 = _ENgetnodetype@8                    
    ENgetnodevalue                = _ENgetnodevalue@12                  
    ENgetnumdemands               = _ENgetnumdemands@8
    ENgetoption                   = _ENgetoption@8                      
    ENgetpatternid                = _ENgetpatternid@8                   
    ENgetpatternindex             = _ENgetpatternindex@8                
    ENgetpatternlen               = _ENgetpatternlen@8                  
    ENgetpatternvalue             = _ENgetpatternvalue@12               
    ENgetpremise                  = _ENgetpremise@36
    ENgetpumptype                 = _ENgetpumptype@8
    ENgetqualinfo                 = _ENgetqualinfo@16
    ENgetqualtype                 = _ENgetqualtype@8
    ENgetresultindex              = _ENgetresultindex@12    
    ENgetrule                     = _ENgetrule@20
    ENgetruleID                   = _ENgetruleID@8
    ENgetstatistic                = _ENgetstatistic@8
    ENgetthenaction               = _ENgetthenaction@20
    ENgettimeparam                = _ENgettimeparam@8
    ENgettitle                    = _ENgettitle@12    
    ENgetversion                  = _ENgetversion@4
    ENgetvertex                   = _ENgetvertex@16
    ENgetvertexcount              = _ENgetvertexcount@8    
    ENinit                        = _ENinit@16
    ENinitH                       = _ENinitH@4                          
    ENinitQ                       = _ENinitQ@4                          
    ENnextH                       = _ENnextH@4                          
    ENnextQ                       = _ENnextQ@4
    ENopen                        = _ENopen@12                          
    ENopenH                       = _ENopenH@0                          
    ENopenQ                       = _ENopenQ@0
    ENreport                      = _ENreport@0                         
    ENresetreport                 = _ENresetreport@0                    
    ENrunH                        = _ENrunH@4                           
    ENrunQ                        = _ENrunQ@4
    ENsaveH                       = _ENsaveH@0                          
    ENsavehydfile                 = _ENsavehydfile@4                    
    ENsaveinpfile                 = _ENsaveinpfile@4                    
    ENsetbasedemand               = _ENsetbasedemand@12
    ENsetcomment                  = _ENsetcomment@12
    ENsetcontrol                  = _ENsetcontrol@24                    
    ENsetcoord                    = _ENsetcoord@20
    ENsetcurve                    = _ENsetcurve@16
    ENsetcurveid                  = _ENsetcurveid@8    
    ENsetcurvevalue               = _ENsetcurvevalue@16               
    ENsetdemandmodel              = _ENsetdemandmodel@16
    ENsetdemandname               = _ENsetdemandname@12
    ENsetdemandpattern            = _ENsetdemandpattern@12
    ENsetelseaction               = _ENsetelseaction@20
    ENsetflowunits                = _ENsetflowunits@4
    ENsetheadcurveindex           = _ENsetheadcurveindex@8
    ENsetjuncdata                 = _ENsetjuncdata@16
    ENsetlinkid                   = _ENsetlinkid@8
    ENsetlinknodes                = _ENsetlinknodes@12
    ENsetlinktype                 = _ENsetlinktype@12
    ENsetlinkvalue                = _ENsetlinkvalue@12
    ENsetnodeid                   = _ENsetnodeid@8                  
    ENsetnodevalue                = _ENsetnodevalue@12                  
    ENsetoption                   = _ENsetoption@8                      
    ENsetpattern                  = _ENsetpattern@12
    ENsetpatternid                = _ENsetpatternid@8    
    ENsetpatternvalue             = _ENsetpatternvalue@12
    ENsetpipedata                 = _ENsetpipedata@20
    ENsetpremise                  = _ENsetpremise@36
    ENsetpremiseindex             = _ENsetpremiseindex@12
    ENsetpremisestatus            = _ENsetpremisestatus@12
    ENsetpremisevalue             = _ENsetpremisevalue@12
    ENsetqualtype                 = _ENsetqualtype@16                   
    ENsetreport                   = _ENsetreport@4                      
    ENsetrulepriority             = _ENsetrulepriority@8
    ENsetstatusreport             = _ENsetstatusreport@4
    ENsettankdata                 = _ENsettankdata@32
    ENsetthenaction               = _ENsetthenaction@20
    ENsettimeparam                = _ENsettimeparam@8
    ENsettitle                    = _ENsettitle@12    
    ENsetvertices                 = _ENsetvertices@16
    ENsolveH                      = _ENsolveH@0                         
    ENsolveQ                      = _ENsolveQ@0                         
    ENstepQ                       = _ENstepQ@4
    ENusehydfile                  = _ENusehydfile@4
    ENwriteline                   = _ENwriteline@4
