#include <Python.h>


PyObject* method_ENopen(PyObject* self, PyObject* args);
PyObject* method_ENopenX(PyObject* self, PyObject* args);
PyObject* method_ENclose(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENaddcontrol(PyObject* self, PyObject* args);
PyObject* method_ENaddcurve(PyObject* self, PyObject* args);
PyObject* method_ENadddemand(PyObject* self, PyObject* args);
PyObject* method_ENaddlink(PyObject* self, PyObject* args);
PyObject* method_ENaddnode(PyObject* self, PyObject* args);
PyObject* method_ENaddpattern(PyObject* self, PyObject* args);
PyObject* method_ENaddrule(PyObject* self, PyObject* args);
PyObject* method_ENclearreport(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENcloseH(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENcloseQ(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENcopyreport(PyObject* self, PyObject* args);
PyObject* method_ENdeletecontrol(PyObject* self, PyObject* args);
PyObject* method_ENdeletecurve(PyObject* self, PyObject* args);
PyObject* method_ENdeletedemand(PyObject* self, PyObject* args);
PyObject* method_ENdeletelink(PyObject* self, PyObject* args);
PyObject* method_ENdeletenode(PyObject* self, PyObject* args);
PyObject* method_ENdeletepattern(PyObject* self, PyObject* args);
PyObject* method_ENdeleterule(PyObject* self, PyObject* args);
PyObject* method_ENgetaveragepatternvalue(PyObject* self, PyObject* args);
PyObject* method_ENgetbasedemand(PyObject* self, PyObject* args);
PyObject* method_ENgetcomment(PyObject* self, PyObject* args);
PyObject* method_ENgetcontrol(PyObject* self, PyObject* args);
PyObject* method_ENgetcoord(PyObject* self, PyObject* args);
PyObject* method_ENgetcount(PyObject* self, PyObject* args);
PyObject* method_ENgetcurve(PyObject* self, PyObject* args);
PyObject* method_ENgetcurveid(PyObject* self, PyObject* args);
PyObject* method_ENgetcurveindex(PyObject* self, PyObject* args);
PyObject* method_ENgetcurvelen(PyObject* self, PyObject* args);
PyObject* method_ENgetcurvetype(PyObject* self, PyObject* args);
PyObject* method_ENgetcurvevalue(PyObject* self, PyObject* args);
PyObject* method_ENgetdemandindex(PyObject* self, PyObject* args);
PyObject* method_ENgetdemandmodel(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENgetdemandname(PyObject* self, PyObject* args);
PyObject* method_ENgetdemandpattern(PyObject* self, PyObject* args);
PyObject* method_ENgetelseaction(PyObject* self, PyObject* args);
PyObject* method_ENgeterror(PyObject* self, PyObject* args);
PyObject* method_ENgetflowunits(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENgetheadcurveindex(PyObject* self, PyObject* args);
PyObject* method_ENgetlinkid(PyObject* self, PyObject* args);
PyObject* method_ENgetlinkindex(PyObject* self, PyObject* args);
PyObject* method_ENgetlinknodes(PyObject* self, PyObject* args);
PyObject* method_ENgetlinktype(PyObject* self, PyObject* args);
PyObject* method_ENgetlinkvalue(PyObject* self, PyObject* args);
PyObject* method_ENgetnodeid(PyObject* self, PyObject* args);
PyObject* method_ENgetnodeindex(PyObject* self, PyObject* args);
PyObject* method_ENgetnodetype(PyObject* self, PyObject* args);
PyObject* method_ENgetnodevalue(PyObject* self, PyObject* args);
PyObject* method_ENgetnumdemands(PyObject* self, PyObject* args);
PyObject* method_ENgetoption(PyObject* self, PyObject* args);
PyObject* method_ENgetpatternid(PyObject* self, PyObject* args);
PyObject* method_ENgetpatternindex(PyObject* self, PyObject* args);
PyObject* method_ENgetpatternlen(PyObject* self, PyObject* args);
PyObject* method_ENgetpatternvalue(PyObject* self, PyObject* args);
PyObject* method_ENgetpremise(PyObject* self, PyObject* args);
PyObject* method_ENgetpumptype(PyObject* self, PyObject* args);
PyObject* method_ENgetqualinfo(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENgetqualtype(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENgetresultindex(PyObject* self, PyObject* args);
PyObject* method_ENgetrule(PyObject* self, PyObject* args);
PyObject* method_ENgetruleID(PyObject* self, PyObject* args);
PyObject* method_ENgetstatistic(PyObject* self, PyObject* args);
PyObject* method_ENgetthenaction(PyObject* self, PyObject* args);
PyObject* method_ENgettimeparam(PyObject* self, PyObject* args);
PyObject* method_ENgettitle(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENgetversion(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENgetvertex(PyObject* self, PyObject* args);
PyObject* method_ENgetvertexcount(PyObject* self, PyObject* args);
PyObject* method_ENinit(PyObject* self, PyObject* args);
PyObject* method_ENinitH(PyObject* self, PyObject* args);
PyObject* method_ENinitQ(PyObject* self, PyObject* args);
PyObject* method_ENnextH(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENnextQ(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENopenH(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENopenQ(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENreport(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENresetreport(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENrunH(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENrunQ(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENsavehydfile(PyObject* self, PyObject* args);
PyObject* method_ENsaveH(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENsaveinpfile(PyObject* self, PyObject* args);
PyObject* method_ENsetbasedemand(PyObject* self, PyObject* args);
PyObject* method_ENsetcomment(PyObject* self, PyObject* args);
PyObject* method_ENsetcontrol(PyObject* self, PyObject* args);
PyObject* method_ENsetcurveid(PyObject* self, PyObject* args);
PyObject* method_ENsetcurve(PyObject* self, PyObject* args);
PyObject* method_ENsetcoord(PyObject* self, PyObject* args);
PyObject* method_ENsetcurvevalue(PyObject* self, PyObject* args);
PyObject* method_ENsetdemandmodel(PyObject* self, PyObject* args);
PyObject* method_ENsetdemandname(PyObject* self, PyObject* args);
PyObject* method_ENsetdemandpattern(PyObject* self, PyObject* args);
PyObject* method_ENsetelseaction(PyObject* self, PyObject* args);
PyObject* method_ENsetflowunits(PyObject* self, PyObject* args);
PyObject* method_ENsetheadcurveindex(PyObject* self, PyObject* args);
PyObject* method_ENsetjuncdata(PyObject* self, PyObject* args);
PyObject* method_ENsetlinkid(PyObject* self, PyObject* args);
PyObject* method_ENsetlinknodes(PyObject* self, PyObject* args);
PyObject* method_ENsetlinktype(PyObject* self, PyObject* args);
PyObject* method_ENsetlinkvalue(PyObject* self, PyObject* args);
PyObject* method_ENsetnodeid(PyObject* self, PyObject* args);
PyObject* method_ENsetnodevalue(PyObject* self, PyObject* args);
PyObject* method_ENsetoption(PyObject* self, PyObject* args);
PyObject* method_ENsetpattern(PyObject* self, PyObject* args);
PyObject* method_ENsetpatternid(PyObject* self, PyObject* args);
PyObject* method_ENsetpatternvalue(PyObject* self, PyObject* args);
PyObject* method_ENsetpipedata(PyObject* self, PyObject* args);
PyObject* method_ENsetpremise(PyObject* self, PyObject* args);
PyObject* method_ENsetpremiseindex(PyObject* self, PyObject* args);
PyObject* method_ENsetpremisevalue(PyObject* self, PyObject* args);
PyObject* method_ENsetpremisestatus(PyObject* self, PyObject* args);
PyObject* method_ENsetqualtype(PyObject* self, PyObject* args);
PyObject* method_ENsetreport(PyObject* self, PyObject* args);
PyObject* method_ENsetrulepriority(PyObject* self, PyObject* args);
PyObject* method_ENsetstatusreport(PyObject* self, PyObject* args);
PyObject* method_ENsettankdata(PyObject* self, PyObject* args);
PyObject* method_ENsetthenaction(PyObject* self, PyObject* args);
PyObject* method_ENsettimeparam(PyObject* self, PyObject* args);
PyObject* method_ENsettitle(PyObject* self, PyObject* args);
PyObject* method_ENsetvertices(PyObject* self, PyObject* args);
PyObject* method_ENsolveH(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENsolveQ(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENstepQ(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENusehydfile(PyObject* self, PyObject* args);
PyObject* method_ENwriteline(PyObject* self, PyObject* args);
PyObject* method_ENgettag(PyObject* self, PyObject* args);
PyObject* method_ENsettag(PyObject* self, PyObject* args);
PyObject* method_ENtimetonextevent(PyObject* self, PyObject* Py_UNUSED(args));
PyObject* method_ENgetnodevalues(PyObject* self, PyObject* args);
PyObject* method_ENgetlinkvalues(PyObject* self, PyObject* args);
PyObject* method_ENsetvertex(PyObject* self, PyObject* args);
PyObject* method_ENloadpatternfile(PyObject* self, PyObject* args);
PyObject* method_ENsetcurvetype(PyObject* self, PyObject* args);
PyObject* method_ENgetcontrolenabled(PyObject* self, PyObject* args);
PyObject* method_ENsetcontrolenabled(PyObject* self, PyObject* args);
PyObject* method_ENgetruleenabled(PyObject* self, PyObject* args);
PyObject* method_ENsetruleenabled(PyObject* self, PyObject* args);